#!/usr/bin/perl

# Copyright (c) 2001-2003, RedIRIS. All Rights Reserved.
#
# You may distribute under the terms of either the GNU General Public License
# or the Artistic License, as specified in the LICENSE file that was shipped
# with this distribution.
package PAPI::Config;

use strict;
use XML::Checker::Parser;
use Data::Dumper;
use Clone qw(clone);
use vars '$Global';
use vars '$Error';

#
sub new {
	my $object = {};
	return bless $object;
}

#
sub print_conf {

	my $gl = shift;
	print STDERR "Configuration:\n" . Dumper($gl);
}

#
sub clone_conf {
	my $gl = shift;

	my $conf2 = new PAPI::Config();
	%{$conf2} = %{ clone($gl) };

	return $conf2;
}

sub flat_conf {

	my $gl = shift;
	my $key;

	my ( $k, $i, $loc, $ser, $last, $h_root );

	$h_root = $gl->{'Global'};

	foreach $key ( keys( %{$gl} ) ) {
		next if ( $key eq 'Global' );
		if ( $gl->{$key}{'Type'} eq 'Server' ) {
			$ser = $gl->{$key};

			#Flat conf for servers
			print "Flating server: $key\n" if ( $ENV{CONFIG_DEVELOP} == 2 );
			flat_local_confs( $h_root, $ser );

			#Flat conf for containers as Location, Directoy and Files.
			#print "dumper ser".Dumper($ser)."\n";
			foreach my $loc ( keys( %{$ser} ) ) {

				#print "Ref ser: $loc <".ref($ser->{$loc}).">\n";
				if (
					( ref( $ser->{$loc} ) eq "HASH" )
					&& (   $ser->{$loc}{'Type'} eq 'Location'
						|| $ser->{$loc}{'Type'} eq 'Directory'
						|| $ser->{$loc}{'Type'} eq 'Files' )
				  )
				{
					print "Flating loc: $loc\n"
					  if ( $ENV{CONFIG_DEVELOP} == 2 );
					flat_local_confs( $ser, $ser->{$loc} );
				}
			}
		}
	}
}

#
sub get_local_conf {
	my $glob_conf = shift;
	my $server    = shift;
	my $serv_id   = shift;

	if ( !defined $glob_conf->{$server} ) {
		if ( !defined $glob_conf->{Main} ) {
			return undef;
		}
		$server = "Main";
	}
	my $last = @{ $glob_conf->{$server} };
	my $i;
	my $confref = -1;
	for ( $i = 1 ; $i < $last ; $i++ ) {
		my $lconf = $glob_conf->{$server}[$i];
		if ( $serv_id eq $lconf->{Serv} ) {
			$confref = $i;
		}
	}
	if ( $confref < 0 ) {
		return undef;
	}
	else {
		return $glob_conf->{$server}[$confref];
	}
}

#
sub trans_conf {

	my $ori  = shift;
	my $dest = new PAPI::Config();
	my $key;
	my $ser;
	my $loc;

	foreach $key ( keys( %{$ori} ) ) {
		$ser = $ori->{$key};
		print "\nTranslating Server: $key\n" if ( $ENV{CONFIG_DEVELOP} == 2 );
		$dest->{$key} = trans_local_conf($ser);

		foreach $loc ( keys( %{$ser} ) ) {
			if (
				( ref( $ser->{$loc} ) eq "HASH" )
				&& (   $ser->{$loc}{'Type'} eq 'Location'
					|| $ser->{$loc}{'Type'} eq 'Directory'
					|| $ser->{$loc}{'Type'} eq 'Files' )
			  )
			{
				print "\nTranslating Location: $loc\n"
				  if ( $ENV{CONFIG_DEVELOP} == 2 );
				$dest->{$key}{$loc} = trans_local_conf( $ser->{$loc} );
			}
		}
	}

	return $dest;
}

# -----------------------------------------
#
#  Load and parse configuration functions
#
# -----------------------------------------

my $Global;
my $charzone = 0;
my $sconf;
my $conf;
my $server;
my $element;
my $level = 0;
my $paux;
my $ffield;
my $furl;
my $fmatch;
my $fid;
my $fbName;
my $fbIndex;
my $fparams;
my $contID;

#
sub load_conf {

	$Global = shift;
	my $file       = shift;
	my $xmlChecker = shift;

	open( CONF, "$file" ) || die "Error opening $file\n";
	my $cont = join( "", <CONF> );
	close(CONF);
	parse_conf( \$cont, $xmlChecker );
}

#
sub parse_conf {
	my $cont       = shift;
	my $xmlChecker = shift;
	my $p1         = '';

	$charzone = 0;
	$level    = 0;
	$Error    = "";

	if ($xmlChecker) {

		# Intance an XML::Checker::Parser
		$p1 = new XML::Checker::Parser(
			ErrorContext    => 1,
			ParseParamEnt   => 1,
			SkipExternalDTD => 1,
			Handlers        => {
				Start => \&handle_start,
				End   => \&handle_end,
				Char  => \&handle_char
			}
		);
	}
	else {

		# Do not check the xml configuration file.
		$p1 = new XML::Parser(
			ErrorContext => 1,
			Handlers     => {
				Start => \&handle_start,
				End   => \&handle_end,
				Char  => \&handle_char
			}
		);
	}

	local $XML::Checker::FAIL = \&my_fail;
	eval { $p1->parse($$cont) };

	if ($@) {
		$Error = "$@";
		return "ERROR";
	}

	return "OK";
}

#
sub my_fail {
	my $code = shift;
	print STDERR "--------------------------------------\n";
	print STDERR "Procesing PAPI configuration file:\n";
	if ( $code < 200 ) {
		XML::Checker::print_error( $code, @_ );
		die;
	}

	#    print STDERR XML::Checker::error_string ($code, @_) if $code < 200;
	XML::Checker::print_error( $code, @_ );
	print STDERR "--------------------------------------\n";
}

#
sub handle_start {
	my $p  = shift;
	my $el = shift;

	if ( $level == 0 ) {
		if ( $el eq "Global" ) {
			$server            = "Global";
			$level             = 1;
			$Global->{$server} = {};
			$sconf             = init_local_conf();
			$sconf->{String}   = "";
			$sconf->{Type}     = $el;

			#push @{ $Global->{$server} },$sconf;
			$Global->{$server} = $sconf;
			$conf = $sconf;
		}
		elsif ( $el eq "Server" ) {
			if ( $level == 1 ) {

				# Seccion server dentro de la seccin Global
				print "Procesando el server de dentro del Global\n";
			}
			$level++;
			my ( $att, $port, $indep );
			while ( $att = shift ) {
				$server = shift if ( $att eq "name" );
				$port   = shift if ( $att eq "port" );
				$indep  = shift if ( $att eq "independent" );
			}
			print "Procesando el server: $server\n";
			$Global->{$server} = {};
			$sconf             = init_local_conf();
			$sconf->{String}   = "";
			$sconf->{Type}     = $el;

			#push @{ $Global->{$server} },$sconf;
			$Global->{$server} = $sconf;
			if ($port) {
				$sconf->{port} = $port;
			}
			else {
				$sconf->{port} = 80;
			}
			$sconf->{'Independent'} = 0;
			$sconf->{'Independent'} = 1
			  if ( ( defined $indep ) && ( $indep eq "true" ) );

			$conf = $sconf;
		}
	}
	elsif ( ( $level == 1 ) || ( $level == 2 ) ) {
		if (
			( $level == 1 )
			&& (   ( $el eq "Location" )
				|| ( $el eq "Directory" )
				|| ( $el eq "Files" ) )
		  )
		{
			my ( $path, $att, $indep, $filtered, $id );
			while ( $att = shift ) {
				$path     = shift if ( $att eq "path" );
				$indep    = shift if ( $att eq "independent" );
				$filtered = shift if ( $att eq "filtered" );
				$id       = shift if ( $att eq "id" );
			}
			$conf = init_local_conf();

			#push @{ $Global->{$server} },$conf;
			## $Global->{Server} contiene un hash con los datos del servidor, ahora por cada
			## Location|directory|File deber tener una estrada ms que ser otro has con los datos
			## del contenedor.
			##
			$contID                     = $id;
			$Global->{$server}{$contID} = $conf;
			$conf->{Type}               = $el;
			$conf->{String}             = $path;
			$conf->{Independent}        = 0;
			$conf->{Independent}        = 1
			  if ( ( defined $indep ) && ( $indep eq "true" ) );
			$conf->{Serv}     = $id;
			$conf->{Filtered} = 0;
			$conf->{Filtered} = 1
			  if ( ( defined $filtered ) && ( $filtered eq "true" ) );
			my $aux = quotemeta($path);
			$conf->{Loc_pattern} = qr/$aux/;
			$level = 2;
		}
		elsif ( $el eq "PAPI_AS" || $el eq "PAPI_SA" ) {
			my ( $att, $id, $url );
			while ( $att = shift ) {
				$url = shift if ( $att eq "url" );
				$id  = shift if ( $att eq "id" );
			}
			$conf->{PAPI_AS}{$id}{url} = $url;
			$charzone                  = 1;
			$element                   = $el;
			$paux                      = $id;
		}
		elsif ( $el eq "Signoff_Location" ) {
			my $s = {};
			my $att;
			while ( $att = shift ) {
				$s->{regex} = shift if ( $att eq "regex" );
				$s->{url}   = shift if ( $att eq "url" );
			}
			push @{ $conf->{$el} }, $s;
		}
		elsif ( $el eq "PAPI_Filter" ) {
			my $s = {};
			my $att;
			while ( $att = shift ) {
				$s->{act} = shift if ( $att eq "action" );
				$s->{reg} = shift if ( $att eq "regex" );
			}
			push @{ $conf->{$el} }, $s;
		}
		elsif ( $el eq "GPoA_Rewrite" ) {
			my $s = {};
			my $att;
			while ( $att = shift ) {
				$s->{poare} = shift if ( $att eq "PoA_regex" );
				$s->{regex} = shift if ( $att eq "src_regex" );
				$s->{value} = shift if ( $att eq "dst" );
			}
			push @{ $conf->{$el} }, $s;
		}
		elsif ( $el eq "HTTP_Auth" ) {
			my $s = {};
			my $att;
			while ( $att = shift ) {
				$s->{br} = shift if ( $att eq "realm" );
				$s->{bu} = shift if ( $att eq "user" );
				$s->{bp} = shift if ( $att eq "password" );
			}
			push @{ $conf->{$el} }, $s;
		}
		elsif ( $el eq "PAPI_Redirect" ) {
			my $s = {};
			my $att;
			while ( $att = shift ) {
				$s->{regex} = shift if ( $att eq "src_regex" );
				$s->{value} = shift if ( $att eq "dst" );
			}
			push @{ $conf->{$el} }, $s;
		}
		elsif ( $el eq "PAPI_Request_Redirect" ) {
			my $s = {};
			my $att;
			while ( $att = shift ) {
				$s->{regex} = shift if ( $att eq "src_regex" );
				$s->{value} = shift if ( $att eq "dst" );
			}
			push @{ $conf->{$el} }, $s;
		}
		elsif ( $el eq "PAPI_Redirect_by_URL" ) {
			my $s = {};
			my $att;
			while ( $att = shift ) {
				$s->{urlregex} = shift if ( $att eq "url" );
				$s->{regex}    = shift if ( $att eq "src_regex" );
				$s->{value}    = shift if ( $att eq "dst" );
			}
			push @{ $conf->{$el} }, $s;
		}
		elsif ( $el eq "PAPI_POST_Redirect" ) {
			my $s = {};
			my $att;
			while ( $att = shift ) {
				$s->{regex} = shift if ( $att eq "src_regex" );
				$s->{value} = shift if ( $att eq "dst" );
			}
			push @{ $conf->{$el} }, $s;
		}
		elsif ( $el eq "PAPI_Header_Redirect" ) {
			my $s = {};
			my $att;
			while ( $att = shift ) {
				$s->{hdregex} = shift if ( $att eq "hd_regex" );
				$s->{regex}   = shift if ( $att eq "src_regex" );
				$s->{value}   = shift if ( $att eq "dst" );
			}
			push @{ $conf->{$el} }, $s;
		}
		elsif ( $el eq "User_Data_Rewrite" ) {
			my ( $att, $re, $va );
			while ( $att = shift ) {
				$re = shift if ( $att eq "src_regex" );
				$va = shift if ( $att eq "dst" );
			}
			$conf->{$el}{regex} = $re;
			$conf->{$el}{value} = $va;
		}
		elsif ( $el eq "Form_Processor" ) {
			my $att;
			undef $furl;
			undef $fmatch;
			undef $fid;
			undef $fbName;
			undef $fbIndex;
			$fparams = {};
			while ( $att = shift ) {
				$furl    = shift if ( $att eq "url" );
				$fmatch  = shift if ( $att eq "match" );
				$fid     = shift if ( $att eq "form_id" );
				$fbName  = shift if ( $att eq "buttonName" );
				$fbIndex = shift if ( $att eq "buttonIndex" );
			}
			$fparams->{match}       = $fmatch  if ( defined $fmatch );
			$fparams->{formName}    = $fid     if ( defined $fid );
			$fparams->{buttonName}  = $fbName  if ( defined $fbName );
			$fparams->{buttonIndex} = $fbIndex if ( defined $fbIndex );
			$fparams->{settings}    = {};
		}
		elsif ( $el eq "Field" ) {
			my $att;
			while ( $att = shift ) {
				$ffield = shift if ( $att eq "name" );
			}
			$charzone = 1;
			$element  = $el;
		}
		else {
			$charzone = 1;
			$element  = $el;
		}
	}
}

#
sub handle_end {
	my $p  = shift;
	my $el = shift;

	if ( ( $el eq "Location" ) || ( $el eq "Directory" ) || ( $el eq "Files" ) )
	{
		$conf  = $sconf;
		$level = 1;
	}
	elsif ( $el eq "Server" ) {
		$level--;
	}
	elsif ( $el eq "Global" ) {
		$level = 0;
	}
	elsif ( $el eq "Form_Processor" ) {
		$conf->{$el}{$furl} = $fparams;
	}
	else {
		$charzone = 0;
	}
}

#
sub handle_char {
	my $p     = shift;
	my $value = shift;

	if ($charzone) {
		if ( $element eq "PAPI_AS" || $element eq "PAPI_SA" ) {
			$conf->{PAPI_AS}{$paux}{desc} = $value;
		}
		elsif ( $element eq "Cookie_Reject" ) {
			push @{ $conf->{$element} }, $value;
		}
		elsif ( $element eq "Field" ) {
			$fparams->{settings}{$ffield} = $value;
		}
		elsif ( $element eq "Reject_POST_Pattern" ) {
			push @{ $conf->{$element} }, $value;
		}
		elsif ( $element eq "Pass_POST_Pattern" ) {
			push @{ $conf->{$element} }, $value;
		}
		elsif ( $element eq "Rewrite_URL_Pattern" ) {
			push @{ $conf->{$element} }, $value;
		}
		elsif ( $element eq "Pass_URL_Pattern" ) {
			push @{ $conf->{$element} }, $value;
		}
		elsif ( $element eq "Reject_URL_Pattern" ) {
			push @{ $conf->{$element} }, $value;
		}
		elsif ( $element eq "Apache_Tags" ) {
			$conf->{$element} .= $value;
		}
		else {
			$conf->{$element} = $value;
		}
	}
}

#
sub trans {

	my $ini_word = shift;

	my %dict = (
		HKey                     => 'Hkey',
		LKey                     => 'Lkey',
		User_Data_Rewrite        => 'RwUser',
		Hash_User_Data           => 'HashUser',
		Remote_URL               => 'Remote_Serv',
		Remote_Domain            => 'Remote_Dom',
		GPoA_Hash_User_Data      => 'GPoAHashUser',
		Client_Address_In_Tokens => 'AddressInToken',
		Redirect_All             => 'RwAll',
		Eval_Proxy_Redirects     => 'EvalPx',
		Proxy_Chunck_Size        => 'PxCkSize',
		Proxy_Server             => 'Proxy',
		Local_IP_Address         => 'PxIP',
		NO_XML                   => 'NoXML',
		Strip_Location           => 'StripLocation',
		Rewrite_MIME_Types       => 'RewMimeTypes',
		Max_TTL                  => 'MaxTtl',
		Attribute_Separator      => 'AttSep',
		Value_Separator          => 'ValSep',

		SPOCP_Server         => 'SPOCPSrv',
		SPOCP_Accept_If_Down => 'SPOCPAcc',
		SPOCP_Action         => 'SPOCPAct',

		Use_Hcook        => 'useHcook',
		Max_Nonce_Errors => 'MxNonceErr',
		Hcook_Generator  => 'HcookGen',

		PADATH_LAA_Handler              => 'PADATH_LAA',
		PADATH_HDD_Handler              => 'PADATH_HDD',
		PADATH_Institution_ID           => 'Athens_IID',
		PADATH_Athens_AuthPoint         => 'Athens_AAP',
		PADATH_Key                      => 'Athens_Key',
		PADATH_Permission_Set_Attribute => 'Athens_psa',
		PADATH_User_Id_Attribute        => 'Athens_uia',
		PADATH_AAP_TTL                  => 'Athens_TTL',

		PAPI_AS               => 'ases',
		PAPI_Filter           => 'filters',
		Cookie_Reject         => 'cookieRejects',
		GPoA_Rewrite          => 'PoARw',
		HTTP_Auth             => 'HttpAuth',
		PAPI_Redirect         => 'redirects',
		PAPI_Redirect_by_URL  => 'url_redirects',
		PAPI_Request_Redirect => 'rqRedirects',
		PAPI_POST_Redirect    => 'postRedirects',
		PAPI_Header_Redirect  => 'hdRedirects',
		Form_Processor        => 'formProc',
		Reject_URL_Pattern    => 'RejUrlPatterns',
		Reject_POST_Pattern   => 'RejPostPatterns',
		Pass_URL_Pattern      => 'PassUrlPatterns',
		Pass_POST_Pattern     => 'PassPostPatterns',
		Rewrite_URL_Pattern   => 'RewUrlPatterns',
		Signoff_Location      => 'signoffLocs'
	);

	if ( !defined( $dict{$ini_word} ) ) {
		return "$ini_word";
	}
	return $dict{$ini_word};
}

# -----------------------------------------
#
#  Local configuration functions
#
# -----------------------------------------

sub init_local_conf {

	my $cnf = {};
	$cnf->{PAPI_AS}               = {};
	$cnf->{PAPI_Filter}           = [];
	$cnf->{Cookie_Reject}         = [];
	$cnf->{GPoA_Rewrite}          = [];
	$cnf->{HTTP_Auth}             = [];
	$cnf->{PAPI_Redirect}         = [];
	$cnf->{PAPI_Redirect_by_URL}  = [];
	$cnf->{PAPI_Request_Redirect} = [];
	$cnf->{PAPI_POST_Redirect}    = [];
	$cnf->{PAPI_Header_Redirect}  = [];
	$cnf->{Form_Processor}        = {};
	$cnf->{Rewrite_URL_Pattern}   = [];
	$cnf->{Pass_URL_Pattern}      = [];
	$cnf->{Reject_URL_Pattern}    = [];
	$cnf->{Pass_POST_Pattern}     = [];
	$cnf->{Reject_POST_Pattern}   = [];
	$cnf->{Signoff_Location}      = [];

	return $cnf;
}

sub flat_local_confs {

	my $father = shift;
	my $son    = shift;

	for my $elem ( keys %{$father} ) {
		if ( $elem eq "PAPI_AS" ) {
			for my $as_id ( keys %{ $father->{$elem} } ) {
				$son->{$elem}{$as_id}{'url'}  = $father->{$elem}{$as_id}{'url'};
				$son->{$elem}{$as_id}{'desc'} =
				  $father->{$elem}{$as_id}{'desc'};
			}
		}
		elsif ( $elem eq "Form_Processor" && ( 0 != %{ $father->{$elem} } ) ) {
			%{ $son->{$elem} } = ( %{ $son->{$elem} }, %{ $father->{$elem} } );
		}
		elsif (( $elem eq "PAPI_Filter" )
			|| ( $elem eq "Cookie_Reject" )
			|| ( $elem eq "GPoA_Rewrite" )
			|| ( $elem eq "HTTP_Auth" )
			|| ( $elem eq "PAPI_Redirect" )
			|| ( $elem eq "PAPI_Redirect_by_URL" )
			|| ( $elem eq "PAPI_Request_Redirect" )
			|| ( $elem eq "Pass_URL_pattern" )
			|| ( $elem eq "Reject_URL_Pattern" )
			|| ( $elem eq "Pass_POST_pattern" )
			|| ( $elem eq "Reject_POST_Pattern" )
			|| ( $elem eq "PAPI_POST_Redirect" )
			|| ( $elem eq "Rewrite_URL_Pattern" )
			|| ( $elem eq "PAPI_Header_Redirect" )
			|| ( $elem eq "Signoff_Location" ) )
		{
			@{ $son->{$elem} } = ( @{ $father->{$elem} }, @{ $son->{$elem} } );
		}
		elsif (( $elem eq "String" )
			|| ( $elem eq "Type" )
			|| ( $elem eq "Serv" )
			|| ( $elem eq "Independent" ) )
		{
			next;
		}
		else {
			if ( !ref( $father->{$elem} ) ) {
				if ( !defined $son->{$elem} ) {
					print "Coping and pasting <$elem>\n"
					  if ( $ENV{CONFIG_DEVELOP} == 2 );
					$son->{$elem} = $father->{$elem};
				}
			}
		}
	}
}

sub trans_local_conf {

	my $ori = shift;

	my $lc = ();

	for my $i ( keys %{$ori} ) {
		if ( $i eq "PAPI_AS" ) {
			print "trans_local_conf# $i a " . trans($i) . "\n"
			  if ( $ENV{CONFIG_DEVELOP} == 2 );
			for my $k ( keys %{ $ori->{$i} } ) {
				$lc->{ trans($i) }{$k}{url}  = $ori->{$i}{$k}{url};
				$lc->{ trans($i) }{$k}{desc} = $ori->{$i}{$k}{desc};
			}
		}
		elsif ( $i eq "Form_Processor" ) {
			print "trans_local_conf# $i a " . trans($i) . "\n"
			  if ( $ENV{CONFIG_DEVELOP} == 2 );
			%{ $lc->{ trans($i) } } = ( %{ $ori->{$i} } );
		}
		elsif (( $i eq "PAPI_Filter" )
			|| ( $i    eq "Cookie_Reject" )
			|| ( $i    eq "GPoA_Rewrite" )
			|| ( $i    eq "HTTP_Auth" )
			|| ( $i    eq "PAPI_Redirect" )
			|| ( $i    eq "PAPI_Redirect_by_URL" )
			|| ( $i    eq "PAPI_Request_Redirect" )
			|| ( $i    eq "Pass_URL_Pattern" )
			|| ( $i    eq "Reject_URL_Pattern" )
			|| ( $i    eq "Pass_POST_Pattern" )
			|| ( $i    eq "Reject_POST_Pattern" )
			|| ( $i    eq "PAPI_POST_Redirect" )
			|| ( $i    eq "Rewrite_URL_Pattern" )
			|| ( $i    eq "PAPI_Header_Redirect" )
			|| ( $i    eq "Signoff_Location" ) )
		{
			print "trans_local_conf# $i a " . trans($i) . "\n"
			  if ( $ENV{CONFIG_DEVELOP} == 2 );
			@{ $lc->{ trans($i) } } = ( @{ $ori->{$i} } );
		}
		else {
			print "trans_local_conf# $i a " . trans($i) . "\n"
			  if ( $ENV{CONFIG_DEVELOP} == 2 );
			$lc->{ trans($i) } = $ori->{$i};
		}

	}
	return $lc;
}

sub print_local_conf {

	my $cf = shift;

	print STDERR "Local Conf: " . Dumper($cf);
}

# ---------------------------------------------------

# my $myconf = new PAPI::Config();

# $myconf->load_conf("/usr/local/apache/conf/papi.conf");
# $myconf->print_conf();
# my $myconf2 = $myconf->clone_conf();
# $myconf2->flat_conf();
# my $c=$myconf2->get_local_conf("localhost.localdomain.es","/manual/trttr");
# print "Conf:" . Dumper($c);

1;

__END__

